@@ -4,7 +4,7 @@
  * (pharvey@codebydesign.com).
  *
  * Modified and extended by Nick Gorham
- * (nick@easysoft.com).
+ * (nick@lurcher.org).
  *
  * Any bugs or problems should be considered the fault of Nick and not
  * Peter.
@@ -321,7 +321,7 @@ extern int pooling_enabled;
 void __generate_connection_string( struct con_struct *con_str, char *str, int str_len )
 {
 struct con_pair *cp;
-char tmp[ 1024 ];
+char *tmp;
 
     str[ 0 ] = '\0';
 
@@ -333,9 +333,28 @@ char tmp[ 1024 ];
     cp = con_str -> list;
     while( cp )
     {
-        if( strcasecmp( cp -> keyword, "DRIVER" ) == 0 )
+        size_t attrlen = strlen( cp -> attribute );
+        int use_esc = isspace( *(cp -> attribute ) ) || attrlen && isspace( cp->attribute[attrlen - 1] );
+        for ( tmp = cp -> attribute; *tmp; tmp++ )
         {
-            sprintf( tmp, "%s={%s};", cp -> keyword, cp -> attribute );
+            use_esc |= (*tmp == '{') || (*tmp == '}');
+            attrlen += (*tmp == '}');
+        }
+        tmp = malloc( strlen( cp -> keyword ) + attrlen + 10 );
+        if( use_esc )
+        {
+            char *tmp2 = tmp + sprintf( tmp, "%s={", cp -> keyword );
+            char *tmp3;
+            for( tmp3 = cp -> attribute; *tmp3 ; tmp3++ )
+            {
+                *tmp2++ = *tmp3;
+                if ( '}' == *tmp3++ )
+                {
+                    *tmp2++ = '}';
+                }
+            }
+            *tmp2++ = '}';
+            *tmp2++ = 0;
         }
         else
         {
@@ -350,6 +369,7 @@ char tmp[ 1024 ];
         {
             strcat( str, tmp );
         }
+        free( tmp );
                                 
         cp = cp -> next;
     }
@@ -362,17 +382,22 @@ int len;
 
     *keyword = *value = NULL;
 
-    ptr = *cp;
+    while ( isspace( **cp ) || **cp == ';' )
+    {
+        (*cp)++;
+    }
 
     if ( !**cp )
         return;
 
+    ptr = *cp;
+
     /* 
      * To handle the case attribute in which the attribute is of the form
      * "ATTR;" instead of "ATTR=VALUE;"
      */
 
-    while ( **cp && **cp != ';' && **cp != '=' )
+    while ( **cp && **cp != '=' )
     {
         (*cp)++;
     }
@@ -385,51 +410,48 @@ int len;
     memcpy( *keyword, ptr, len );
     (*keyword)[ len ] = '\0';
 
-    if (**cp != ';') {
 		(*cp)++;
-	}
 
-    ptr = *cp;
-
-    if ( strcmp( *keyword, "DRIVER" ) == 0 )
+    if ( **cp == '{' )
     {
-        if ( **cp && **cp == '{' )
+        /* escaped with '{' - all characters until next '}' not followed by '}', or
+           end of string, are part of the value */
+        int i = 0;
+        ptr = ++*cp;
+        while ( **cp && (**cp != '}' || (*cp)[1] == '}') )
         {
+            if ( **cp == '}' )
             (*cp)++;
-            ptr ++;
-            while ( **cp && **cp != '}' )
                 (*cp)++;
-
-            len = *cp - ptr;
-            *value = malloc( len + 1 );
-            memcpy( *value, ptr , len );
-            (*value)[ len ] = '\0';
-            (*cp)++;
         }
-        else
-        {
-            while ( **cp && **cp != ';' )
-                (*cp)++;
-
             len = *cp - ptr;
             *value = malloc( len + 1 );
-            memcpy( *value, ptr, len );
-            (*value)[ len ] = '\0';
+        while( ptr < *cp )
+        {
+            if ( ((*value)[i++] = *ptr++) == '}')
+            {
+                ptr++;
+        }
+        }
+        (*value)[i] = 0;
+        if ( **cp == '}' )
+        {
+                (*cp)++;
         }
     }
     else
     {
+        /* non-escaped: all characters until ';' or end of string are value */
+        ptr = *cp;
         while ( **cp && **cp != ';' )
+        {
             (*cp)++;
-
+        }
         len = *cp - ptr;
         *value = malloc( len + 1 );
         memcpy( *value, ptr, len );
-        (*value)[ len ] = '\0';
+        (*value)[ len ] = 0;
     }
-
-    if ( **cp )
-        (*cp)++;
 }
 
 struct con_pair * __get_pair( char ** cp )
@@ -463,7 +485,7 @@ struct con_pair *ptr, *end;
         ptr = con_str -> list;
         while( ptr )
         {
-            if( strcmp( kword, ptr -> keyword ) == 0 )
+            if( strcasecmp( kword, ptr -> keyword ) == 0 )
             {
                 free( ptr -> attribute );
                 ptr -> attribute = malloc( strlen( value ) + 1 );
@@ -536,16 +558,26 @@ int got_driver = 0;    /* if we have a DRIVER or FILEDSN then ignore any DSN */
     {
         if ( strcasecmp( cp -> keyword, "DSN" ) == 0 )
         {
-            if ( got_driver && exclude )
+            if ( got_driver && exclude ) {
+                /* 11-29-2010 JM Modify to free the allocated memory before continuing. */
+                free( cp -> keyword );
+                free( cp -> attribute );
+                free( cp );
                 continue;
+            }
 
             got_dsn = 1;
         }
         else if ( strcasecmp( cp -> keyword, "DRIVER" ) == 0 ||
-            strcmp( cp -> keyword, "FILEDSN" ) == 0 )
+            strcasecmp( cp -> keyword, "FILEDSN" ) == 0 )
         {
-            if ( got_dsn && exclude )
+            if ( got_dsn && exclude ) {
+                /* 11-29-2010 JM Modify to free the allocated memory before continuing. */
+                free( cp -> keyword );
+                free( cp -> attribute );
+                free( cp );
                 continue;
+            }
 
             got_driver = 1;
         }
@@ -607,6 +639,21 @@ void __release_conn( struct con_struct *con_str )
     con_str -> count = 0;
 }
 
+void __handle_attr_extensions_cs( DMHDBC connection, struct con_struct *con_str )
+{
+    char *ptr;
+
+    if (( ptr = __get_attribute_value( con_str, "DMEnvAttr" )) != NULL ) {
+        __parse_attribute_string( &connection -> env_attribute, ptr, SQL_NTS );
+    }
+    if (( ptr = __get_attribute_value( con_str, "DMConnAttr" )) != NULL ) {
+        __parse_attribute_string( &connection -> dbc_attribute, ptr, SQL_NTS );
+    }
+    if (( ptr = __get_attribute_value( con_str, "DMStmtAttr" )) != NULL ) {
+        __parse_attribute_string( &connection -> stmt_attribute, ptr, SQL_NTS );
+    }
+}
+
 SQLRETURN SQLDriverConnectA(
     SQLHDBC            hdbc,
     SQLHWND            hwnd,
@@ -639,15 +686,15 @@ SQLRETURN SQLDriverConnect(
 {
     DMHDBC connection = (DMHDBC)hdbc;
     struct con_struct con_struct;
-    char *driver, *dsn = NULL, *filedsn, *tsavefile, savefile[ 128 ];
+    char *driver, *dsn = NULL, *filedsn, *tsavefile, savefile[ INI_MAX_PROPERTY_VALUE + 1 ];
     char lib_name[ INI_MAX_PROPERTY_VALUE + 1 ];
     char driver_name[ INI_MAX_PROPERTY_VALUE + 1 ];
     SQLRETURN ret_from_connect;
     SQLCHAR s1[ 2048 ];
     SQLCHAR local_conn_str_in[ 2048 ];
     SQLCHAR local_out_conection[ 2048 ];
-    char save_filedsn[ 128 ];
-    int warnings;
+    char *save_filedsn;
+    int warnings = 0;
 
     /*
      * check connection
@@ -818,22 +865,22 @@ SQLRETURN SQLDriverConnect(
 			else 
 			{
 				prefix = returned_dsn;
-				target = strchr( returned_dsn, '=' );
+				target = (SQLCHAR*)strchr( (char*)returned_dsn, '=' );
 				if ( target ) 
 				{
 					*target = '\0';
 					target ++;
-        			__append_pair( &con_struct, prefix, target );
+        			__append_pair( &con_struct, (char*)prefix, (char*)target );
 				}
 				else {
-        			__append_pair( &con_struct, "DSN", returned_dsn );
+        			__append_pair( &con_struct, "DSN", (char*)returned_dsn );
 				}
 			}
 
 			/*
 			 * regenerate to pass to driver
 			 */
-			__generate_connection_string( &con_struct, local_conn_str_in, sizeof( local_conn_str_in ));
+			__generate_connection_string( &con_struct, (char*)local_conn_str_in, sizeof( local_conn_str_in ));
 			conn_str_in = local_conn_str_in;
         	len_conn_str_in = strlen((char*) conn_str_in );
 		}
@@ -944,7 +991,13 @@ SQLRETURN SQLDriverConnect(
     tsavefile = __get_attribute_value( &con_struct, "SAVEFILE" );
     if ( tsavefile )
     {
-        strcpy( savefile, tsavefile );
+        if ( strlen( tsavefile ) > INI_MAX_PROPERTY_VALUE ) {
+            memcpy( savefile, tsavefile, INI_MAX_PROPERTY_VALUE );
+            savefile[ INI_MAX_PROPERTY_VALUE ] = '\0';
+        }
+        else {
+            strcpy( savefile, tsavefile );
+        }
     }
     else
     {
@@ -952,20 +1005,20 @@ SQLRETURN SQLDriverConnect(
     }
 
     /*
-     * open the file dsn, get each entry from it, if its not in the connection 
+     * open the file dsn, get each entry from it, if it's not in the connection 
      * struct, add it
      */
 
     filedsn = __get_attribute_value( &con_struct, "FILEDSN" );
     if ( filedsn )
     {
-        char str[ 2048 ];
+        char str[ 1024 * 16 ];
 
         if ( SQLReadFileDSN( filedsn, "ODBC", NULL, str, sizeof( str ), NULL ))
         {
             struct con_struct con_struct1;
 
-	    	strcpy( save_filedsn, filedsn );
+            save_filedsn = strdup( filedsn );
 
             if ( strlen( str ))
             {
@@ -988,15 +1041,35 @@ SQLRETURN SQLDriverConnect(
                     cp = con_struct.list;
                     while( cp )
                     {
-		    			char str1[ 256 ];
+		    			char *str1;
 
 						/*
 			 			* Don't pass FILEDSN down
 			 			*/
 
-						if ( strcmp( cp -> keyword, "FILEDSN" ) &&
-							strcmp( cp -> keyword, "FILEDSN" ) )
+						if ( strcasecmp( cp -> keyword, "FILEDSN" ) &&
+							strcasecmp( cp -> keyword, "FILEDSN" ) )
 						{
+                            str1 = malloc( strlen( cp -> keyword ) + strlen( cp -> attribute ) + 10 );
+
+                            if ( !str1 ) {
+                                dm_log_write( __FILE__, 
+                                        __LINE__, 
+                                        LOG_INFO, 
+                                        LOG_INFO, 
+                                        "Error: HY001" );
+                        
+                                __post_internal_error( &connection -> error,
+                                        ERROR_HY001, NULL,
+                                        connection -> environment -> requested_version );
+
+                                if ( save_filedsn ) {
+                                    free( save_filedsn );
+                                }
+                        
+                                return function_return( SQL_HANDLE_DBC, connection, SQL_ERROR );
+                            }
+
 							if ( strlen((char*) conn_str_in ) > 0 )
 							{
                 				sprintf( str1, ";%s=%s", cp -> keyword, cp -> attribute );
@@ -1005,7 +1078,18 @@ SQLRETURN SQLDriverConnect(
 							{
                    				sprintf( str1, "%s=%s", cp -> keyword, cp -> attribute );
 							}
-                			strcat((char*) conn_str_in, str1 );
+
+                            if ( strlen( (char*)conn_str_in ) + strlen( str1 ) < conn_str_out_max ) {
+                			    strcat((char*) conn_str_in, str1 );
+                            }
+                            else {
+                                warnings = 1;
+                                __post_internal_error( &connection -> error,
+                                        ERROR_01004, NULL,
+                                        connection -> environment -> requested_version );
+                            }
+
+                            free( str1 );
 						}
 
                         cp = cp -> next;
@@ -1021,7 +1105,26 @@ SQLRETURN SQLDriverConnect(
                     {
                         if ( !__get_attribute_value( &con_struct, cp -> keyword ))
                         {
-                            char str1[ 256 ];
+                            char *str1;
+
+                            str1 = malloc( strlen( cp -> keyword ) + strlen( cp -> attribute ) + 10 );
+                            if ( !str1 ) {
+                                dm_log_write( __FILE__, 
+                                        __LINE__, 
+                                        LOG_INFO, 
+                                        LOG_INFO, 
+                                        "Error: HY001" );
+                        
+                                __post_internal_error( &connection -> error,
+                                        ERROR_HY001, NULL,
+                                        connection -> environment -> requested_version );
+
+                                if ( save_filedsn ) {
+                                    free( save_filedsn );
+                                }
+                        
+                                return function_return( SQL_HANDLE_DBC, connection, SQL_ERROR );
+                            }
 
 							if ( strlen((char*) conn_str_in ) > 0 )
 							{
@@ -1031,7 +1134,16 @@ SQLRETURN SQLDriverConnect(
 							{
                             	sprintf( str1, "%s=%s", cp -> keyword, cp -> attribute );
 							}
-                            strcat((char*) conn_str_in, str1 );
+                            if ( strlen( (char*)conn_str_in ) + strlen( str1 ) < conn_str_out_max ) {
+                                strcat((char*) conn_str_in, str1 );
+                            }
+                            else {
+                                warnings = 1;
+                                __post_internal_error( &connection -> error,
+                                        ERROR_01004, NULL,
+                                        connection -> environment -> requested_version );
+                            }
+                            free( str1 );
                         }
                         cp = cp -> next;
                     }
@@ -1051,10 +1163,14 @@ SQLRETURN SQLDriverConnect(
             __parse_connection_string( &con_struct,
                     (char*)conn_str_in, len_conn_str_in );
         }
+        else
+        {
+            save_filedsn = NULL;
+        }
     }
     else
     {
-	    save_filedsn[ 0 ] = '\0';
+	    save_filedsn = NULL;
     }
 
     /*
@@ -1062,7 +1178,6 @@ SQLRETURN SQLDriverConnect(
      *
      * have we got a DRIVER= attribute
      */
-
     driver = __get_attribute_value( &con_struct, "DRIVER" );
     if ( driver )
     {
@@ -1070,6 +1185,24 @@ SQLRETURN SQLDriverConnect(
          * look up the driver in the ini file
          */
 
+        if ( strlen( driver ) >= sizeof( driver_name )) {
+            dm_log_write( __FILE__, 
+                    __LINE__, 
+                    LOG_INFO, 
+                    LOG_INFO, 
+                    "Error: IM011" );
+                        
+            __post_internal_error( &connection -> error,
+                    ERROR_IM011, NULL,
+                    connection -> environment -> requested_version );
+
+            if ( save_filedsn ) {
+                free( save_filedsn );
+            }
+                        
+            return function_return( SQL_HANDLE_DBC, connection, SQL_ERROR );
+        }
+
         strcpy( driver_name, driver );
 
 #ifdef PLATFORM64
@@ -1086,26 +1219,12 @@ SQLRETURN SQLDriverConnect(
                 lib_name, sizeof( lib_name ), "ODBCINST.INI" );
 #endif
 
-        if ( lib_name[ 0 ] == '\0' )
-        {
-            /*
-             * at this point a box could pop up to allow the selection of a driver
-             *
-             * do this later
-             */
+        /*
+         * Assume if it's not in a odbcinst.ini then it's a direct reference
+         */
 
-            dm_log_write( __FILE__, 
-                    __LINE__, 
-                    LOG_INFO, 
-                    LOG_INFO, 
-                    "Error: IM002" );
-
-            __post_internal_error( &connection -> error,
-                    ERROR_IM002, NULL,
-                    connection -> environment -> requested_version );
-            __release_conn( &con_struct );
-
-            return function_return( SQL_HANDLE_DBC, connection, SQL_ERROR );
+        if ( lib_name[ 0 ] == '\0' ) {
+            strcpy( lib_name, driver );
         }
 
         strcpy( connection -> dsn, "" );
@@ -1128,6 +1247,10 @@ SQLRETURN SQLDriverConnect(
 
             __release_conn( &con_struct );
 
+            if ( save_filedsn ) {
+                free( save_filedsn );
+            }
+
             return function_return( SQL_HANDLE_DBC, connection, SQL_ERROR );
         }
 
@@ -1145,6 +1268,10 @@ SQLRETURN SQLDriverConnect(
 
             __release_conn( &con_struct );
 
+            if ( save_filedsn ) {
+                free( save_filedsn );
+            }
+
             return function_return( SQL_HANDLE_DBC, connection, SQL_ERROR );
         }
 
@@ -1166,6 +1293,10 @@ SQLRETURN SQLDriverConnect(
 
             __release_conn( &con_struct );
 
+            if ( save_filedsn ) {
+                free( save_filedsn );
+            }
+
             return function_return( SQL_HANDLE_DBC, connection, SQL_ERROR );
         }
 
@@ -1182,6 +1313,12 @@ SQLRETURN SQLDriverConnect(
 
         __handle_attr_extensions( connection, dsn, driver_name );
     }
+    else {
+        /* 
+         * the attributes may be in the connection string
+         */
+        __handle_attr_extensions_cs( connection, &con_struct );
+    }
 
     __release_conn( &con_struct );
 
@@ -1190,6 +1327,12 @@ SQLRETURN SQLDriverConnect(
      */
     if ( !__connect_part_one( connection, lib_name, driver_name, &warnings ))
     {
+        if ( save_filedsn ) {
+            free( save_filedsn );
+        }
+
+        __disconnect_part_four( connection );       /* release unicode handles */
+
         return function_return( SQL_HANDLE_DBC, connection, SQL_ERROR );
     }
 
@@ -1203,10 +1346,15 @@ SQLRETURN SQLDriverConnect(
                 "Error: IM001" );
 
         __disconnect_part_one( connection );
+        __disconnect_part_four( connection );       /* release unicode handles */
         __post_internal_error( &connection -> error,
                 ERROR_IM001, NULL,
                 connection -> environment -> requested_version );
 
+        if ( save_filedsn ) {
+            free( save_filedsn );
+        }
+
         return function_return( SQL_HANDLE_DBC, connection, SQL_ERROR );
     }
 
@@ -1264,7 +1412,7 @@ SQLRETURN SQLDriverConnect(
 
                     if ( SQL_SUCCEEDED( ret ))
                     {
-                        __post_internal_error_ex( &connection -> error,
+                        __post_internal_error_ex_noprefix( &connection -> error,
                                 sqlstate,
                                 native_error,
                                 message_text,
@@ -1297,7 +1445,7 @@ SQLRETURN SQLDriverConnect(
 
                     if ( SQL_SUCCEEDED( ret ))
                     {
-                        __post_internal_error_ex( &connection -> error,
+                        __post_internal_error_ex_noprefix( &connection -> error,
                                 sqlstate,
                                 native_error,
                                 message_text,
@@ -1333,6 +1481,10 @@ SQLRETURN SQLDriverConnect(
                     LOG_INFO,
                     connection -> msg );
 
+            if ( save_filedsn ) {
+                free( save_filedsn );
+            }
+
             return function_return( SQL_HANDLE_DBC, connection, ret_from_connect );
         }
 		connection -> unicode_driver = 0;
@@ -1341,8 +1493,10 @@ SQLRETURN SQLDriverConnect(
     {
         SQLWCHAR *uc_conn_str_in, *s1 = NULL;
         SQLCHAR s2[ 128 ];
+        int wlen;
 
-        uc_conn_str_in = ansi_to_unicode_alloc( conn_str_in, len_conn_str_in, connection );
+        uc_conn_str_in = ansi_to_unicode_alloc( conn_str_in, len_conn_str_in, connection, &wlen );
+        len_conn_str_in = wlen;
 
         if ( CHECK_SQLSETCONNECTATTR( connection ))
         {
@@ -1405,14 +1559,14 @@ SQLRETURN SQLDriverConnect(
                     {
                         SQLCHAR *as1, *as2;
 
-                        __post_internal_error_ex_w( &connection -> error,
+                        __post_internal_error_ex_w_noprefix( &connection -> error,
                                 sqlstate,
                                 native_error,
                                 message_text,
                                 SUBCLASS_ODBC, SUBCLASS_ODBC );
 
-                        as1 = (SQLCHAR*) unicode_to_ansi_alloc( sqlstate, SQL_NTS, connection );
-                        as2 = (SQLCHAR*) unicode_to_ansi_alloc( message_text, SQL_NTS, connection );
+                        as1 = (SQLCHAR*) unicode_to_ansi_alloc( sqlstate, SQL_NTS, connection, NULL );
+                        as2 = (SQLCHAR*) unicode_to_ansi_alloc( message_text, SQL_NTS, connection, NULL );
 
                         sprintf( connection -> msg, "\t\tDIAG [%s] %s",
                                 as1, as2 );
@@ -1446,14 +1600,14 @@ SQLRETURN SQLDriverConnect(
                     {
                         SQLCHAR *as1, *as2;
 
-                        __post_internal_error_ex_w( &connection -> error,
+                        __post_internal_error_ex_w_noprefix( &connection -> error,
                                 sqlstate,
                                 native_error,
                                 message_text,
                                 SUBCLASS_ODBC, SUBCLASS_ODBC );
 
-                        as1 = (SQLCHAR*) unicode_to_ansi_alloc( sqlstate, SQL_NTS, connection );
-                        as2 = (SQLCHAR*) unicode_to_ansi_alloc( message_text, SQL_NTS, connection );
+                        as1 = (SQLCHAR*) unicode_to_ansi_alloc( sqlstate, SQL_NTS, connection, NULL );
+                        as2 = (SQLCHAR*) unicode_to_ansi_alloc( message_text, SQL_NTS, connection, NULL );
 
                         sprintf( connection -> msg, "\t\tDIAG [%s] %s",
                             as1, as2 );
@@ -1487,13 +1641,22 @@ SQLRETURN SQLDriverConnect(
                     LOG_INFO,
                     connection -> msg );
 
+            if ( save_filedsn ) {
+                free( save_filedsn );
+            }
+
+            if ( s1 )
+            {
+                free( s1 );
+            }
+
             return function_return( SQL_HANDLE_DBC, connection, ret_from_connect );
         }
         else
         {
             if ( conn_str_out && s1 )
             {
-                unicode_to_ansi_copy((char*) conn_str_out, conn_str_out_max, s1, SQL_NTS, connection );
+                unicode_to_ansi_copy((char*) conn_str_out, conn_str_out_max, s1, SQL_NTS, connection, NULL );
             }
         }
 
@@ -1533,6 +1696,10 @@ SQLRETURN SQLDriverConnect(
         __disconnect_part_one( connection );
         __disconnect_part_four( connection );
 
+        if ( save_filedsn ) {
+            free( save_filedsn );
+        }
+
         return function_return( SQL_HANDLE_DBC, connection, SQL_ERROR );
     }
 
@@ -1583,7 +1750,7 @@ SQLRETURN SQLDriverConnect(
 	    }
     }
 
-    if ( strlen( save_filedsn ))
+    if ( save_filedsn && strlen( save_filedsn ))
     {
 	    char *str = strdup((char*) conn_str_out );
 	    strcpy((char*) conn_str_out, "FILEDSN=" );
@@ -1598,6 +1765,10 @@ SQLRETURN SQLDriverConnect(
 	    }
     }
 
+    if ( save_filedsn ) {
+        free( save_filedsn );
+    }
+
     /*
      * write the connection string out to a file
      */
@@ -1671,5 +1842,5 @@ SQLRETURN SQLDriverConnect(
         ret_from_connect = SQL_SUCCESS_WITH_INFO;
     }
 
-    return function_return( SQL_HANDLE_DBC, connection, ret_from_connect );
+    return function_return_nodrv( SQL_HANDLE_DBC, connection, ret_from_connect );
 }
