@@ -4,7 +4,7 @@
  * (pharvey@codebydesign.com).
  *
  * Modified and extended by Nick Gorham
- * (nick@easysoft.com).
+ * (nick@lurcher.org).
  *
  * Any bugs or problems should be considered the fault of Nick and not
  * Peter.
@@ -180,7 +180,7 @@ SQLRETURN SQLEndTran( SQLSMALLINT handle_type,
                     ERROR_HY092, NULL,
                     statement -> connection -> environment -> requested_version );
 
-            return function_return( SQL_HANDLE_STMT, statement, SQL_ERROR );
+            return function_return_nodrv( SQL_HANDLE_STMT, statement, SQL_ERROR );
         }
         else if ( handle_type == SQL_HANDLE_DESC ) {
             if ( !__validate_desc(( DMHDESC ) handle ))
@@ -208,7 +208,7 @@ SQLRETURN SQLEndTran( SQLSMALLINT handle_type,
                     ERROR_HY092, NULL,
                     descriptor -> connection -> environment -> requested_version );
 
-            return function_return( SQL_HANDLE_DESC, descriptor, SQL_ERROR );
+            return function_return_nodrv( SQL_HANDLE_DESC, descriptor, SQL_ERROR );
         }
         else {
              dm_log_write( __FILE__, 
@@ -270,7 +270,7 @@ SQLRETURN SQLEndTran( SQLSMALLINT handle_type,
                     ERROR_HY012, NULL,
                     environment -> requested_version );
 
-            return function_return( SQL_HANDLE_ENV, environment, SQL_ERROR );
+            return function_return_nodrv( SQL_HANDLE_ENV, environment, SQL_ERROR );
         }
 
         if ( environment -> state == STATE_E2 )
@@ -286,11 +286,7 @@ SQLRETURN SQLEndTran( SQLSMALLINT handle_type,
                 if ( connection -> environment == environment &&
                         connection -> state > STATE_C4 )
                 {
-                    if( __check_stmt_from_dbc( connection, STATE_S8 ) ||
-                        __check_stmt_from_dbc( connection, STATE_S9 ) ||
-                        __check_stmt_from_dbc( connection, STATE_S10 ) ||
-                        __check_stmt_from_dbc( connection, STATE_S11 ) ||
-                        __check_stmt_from_dbc( connection, STATE_S12 )) {
+                    if( __check_stmt_from_dbc_v( connection, 8, STATE_S8, STATE_S9, STATE_S10, STATE_S11, STATE_S12, STATE_S13, STATE_S14, STATE_S15 )) {
 
                         dm_log_write( __FILE__, 
                                 __LINE__, 
@@ -302,7 +298,7 @@ SQLRETURN SQLEndTran( SQLSMALLINT handle_type,
                                 ERROR_HY010, NULL,
                                 environment -> requested_version );
             
-                        return function_return( SQL_HANDLE_ENV, environment, SQL_ERROR );
+                        return function_return_nodrv( SQL_HANDLE_ENV, environment, SQL_ERROR );
                     }
                 }
 
@@ -376,7 +372,7 @@ SQLRETURN SQLEndTran( SQLSMALLINT handle_type,
                                 ERROR_IM001, NULL,
                                 environment -> requested_version );
 
-                        return function_return( SQL_HANDLE_ENV, environment, SQL_ERROR );
+                        return function_return_nodrv( SQL_HANDLE_ENV, environment, SQL_ERROR );
                     }
                 }
 
@@ -409,18 +405,20 @@ SQLRETURN SQLEndTran( SQLSMALLINT handle_type,
 
         function_entry( connection );
 
-        sprintf( connection -> msg, "\n\t\tEntry:\
-            \n\t\t\tConnection = %p\
-            \n\t\t\tCompletion Type = %d",
-                (void*)connection,
-                (int)completion_type );
-
-        dm_log_write( __FILE__, 
-                __LINE__, 
-                LOG_INFO, 
-                LOG_INFO, 
-                connection -> msg );
-
+        if ( log_info.log_flag )
+        {
+            sprintf( connection -> msg, "\n\t\tEntry:\
+                \n\t\t\tConnection = %p\
+                \n\t\t\tCompletion Type = %d",
+                    (void*)connection,
+                    (int)completion_type );
+    
+            dm_log_write( __FILE__, 
+                    __LINE__, 
+                    LOG_INFO, 
+                    LOG_INFO, 
+                    connection -> msg );
+        }
         thread_protect( SQL_HANDLE_DBC, connection );
 
         if ( connection -> state == STATE_C1 ||
@@ -437,18 +435,14 @@ SQLRETURN SQLEndTran( SQLSMALLINT handle_type,
                     ERROR_08003, NULL,
                     connection -> environment -> requested_version );
 
-            return function_return( SQL_HANDLE_DBC, connection, SQL_ERROR );
+            return function_return_nodrv( SQL_HANDLE_DBC, connection, SQL_ERROR );
         }
 
         /*
          * check status of statements belonging to this connection
          */
 
-        if( __check_stmt_from_dbc( connection, STATE_S8 ) ||
-            __check_stmt_from_dbc( connection, STATE_S9 ) ||
-            __check_stmt_from_dbc( connection, STATE_S10 ) ||
-            __check_stmt_from_dbc( connection, STATE_S11 ) ||
-            __check_stmt_from_dbc( connection, STATE_S12 )) {
+        if( __check_stmt_from_dbc_v( connection, 8, STATE_S8, STATE_S9, STATE_S10, STATE_S11, STATE_S12, STATE_S13, STATE_S14, STATE_S15 )) {
 
             dm_log_write( __FILE__, 
                     __LINE__, 
@@ -460,7 +454,7 @@ SQLRETURN SQLEndTran( SQLSMALLINT handle_type,
                     ERROR_HY010, NULL,
                     connection -> environment -> requested_version );
 
-            return function_return( SQL_HANDLE_DBC, connection, SQL_ERROR );
+            return function_return_nodrv( SQL_HANDLE_DBC, connection, SQL_ERROR );
         }
 
         if ( completion_type != SQL_COMMIT &&
@@ -476,7 +470,7 @@ SQLRETURN SQLEndTran( SQLSMALLINT handle_type,
                     ERROR_HY012, NULL,
                     connection -> environment -> requested_version );
 
-            return function_return( SQL_HANDLE_DBC, connection, SQL_ERROR );
+            return function_return_nodrv( SQL_HANDLE_DBC, connection, SQL_ERROR );
         }
 
         if ( CHECK_SQLENDTRAN( connection ))
@@ -505,16 +499,14 @@ SQLRETURN SQLEndTran( SQLSMALLINT handle_type,
                     ERROR_IM001, NULL,
                     connection -> environment -> requested_version );
 
-            return function_return( SQL_HANDLE_DBC, connection, SQL_ERROR );
+            return function_return_nodrv( SQL_HANDLE_DBC, connection, SQL_ERROR );
         }
 
 	    if( SQL_SUCCEEDED(ret) )
 	    {
-	    DMHSTMT statement;
-	    SQLINTEGER stmt_remaining;
-	    SQLSMALLINT cb_value;
-	    SQLSMALLINT cb_value_length = sizeof(SQLSMALLINT);
-	    SQLRETURN ret1;
+            SQLSMALLINT cb_value;
+            SQLSMALLINT cb_value_length = sizeof(SQLSMALLINT);
+            SQLRETURN ret1;
 	    
             /*
              * for each statement belonging to this connection set its state 
@@ -560,61 +552,7 @@ SQLRETURN SQLEndTran( SQLSMALLINT handle_type,
 
             if( connection -> cbs_found )
             {
-	    		/* 
-	     	 	 * We need to protect this at a higher level than connection
-	     	 	 * as statements can be coming and going
-	     	 	 */
-	
-	    		mutex_lib_entry();
-	
-            	statement = __get_stmt_root();
-            	stmt_remaining = connection -> statement_count;
-            
-                while ( statement && stmt_remaining > 0 )
-                {
-                    if ( statement -> connection == connection )
-                    {
-                        if ( (statement -> state == STATE_S2 ||
-                              statement -> state == STATE_S3) &&
-                             cb_value == SQL_CB_DELETE )
-                        {
-                            statement -> state = STATE_S1;
-                            statement -> prepared = 0;
-                        }
-                        else if ( statement -> state == STATE_S4 ||
-                              statement -> state == STATE_S5 ||
-                              statement -> state == STATE_S6 ||
-                              statement -> state == STATE_S7 )
-                        {
-                            if( !statement -> prepared && 
-                                (cb_value == SQL_CB_DELETE ||
-                                 cb_value == SQL_CB_CLOSE) )
-                            {
-                                statement -> state = STATE_S1;
-                            }
-                            else if( statement -> prepared )
-                            {
-                                if( cb_value == SQL_CB_DELETE )
-                                {
-                                    statement -> state = STATE_S1;
-                                    statement -> prepared = 0;
-                                }
-                                else if( cb_value == SQL_CB_CLOSE )
-                                {
-                                    if ( statement -> state == STATE_S4 )
-                                      statement -> state = STATE_S2;
-                                    else
-                                      statement -> state = STATE_S3;
-                                }
-                            }
-                        }
-
-                        stmt_remaining --;
-                    }
-
-                    statement = statement -> next_class_list;
-                }
-				mutex_lib_exit();
+                __set_stmt_state( connection, cb_value );
             }
         }
 
