@@ -4,7 +4,7 @@
  * (pharvey@codebydesign.com).
  *
  * Modified and extended by Nick Gorham
- * (nick@easysoft.com).
+ * (nick@lurcher.org).
  *
  * Any bugs or problems should be considered the fault of Nick and not
  * Peter.
@@ -101,6 +101,27 @@ static char const rcsid[]= "$RCSfile: SQLGetDiagFieldW.c,v $";
 HY099,HY100,HY101,HY105,HY107,HY109,HY110,HY111,HYT00,HYT01,IM001,IM002,IM003,\
 IM004,IM005,IM006,IM007,IM008,IM010,IM011,IM012"
 
+/*
+ * is it a diag identifier that we have to convert from unicode to ansi
+ */
+
+static int is_char_diag( int diag_identifier )
+{
+    switch( diag_identifier ) {
+        case SQL_DIAG_CLASS_ORIGIN:
+        case SQL_DIAG_CONNECTION_NAME:
+        case SQL_DIAG_DYNAMIC_FUNCTION:
+        case SQL_DIAG_MESSAGE_TEXT:
+        case SQL_DIAG_SERVER_NAME:
+        case SQL_DIAG_SQLSTATE:
+        case SQL_DIAG_SUBCLASS_ORIGIN:
+            return 1;
+
+        default:
+            return 0;
+    }
+}
+
 static SQLRETURN extract_sql_error_field_w( EHEAD *head,
                 SQLSMALLINT rec_number,
                 SQLSMALLINT diag_identifier,
@@ -110,6 +131,11 @@ static SQLRETURN extract_sql_error_field_w( EHEAD *head,
 {
     ERROR *ptr;
 
+    if ( is_char_diag( diag_identifier ) && buffer_length < 0 )
+    {
+        return SQL_ERROR;
+    }
+
     /*
      * check the header fields first
      */
@@ -122,7 +148,7 @@ static SQLRETURN extract_sql_error_field_w( EHEAD *head,
             SQLINTEGER val;
             SQLRETURN ret;
 
-            if ( head -> handle_type != SQL_HANDLE_STMT )
+            if ( rec_number > 0 || head -> handle_type != SQL_HANDLE_STMT )
             {
                 return SQL_ERROR;
             }
@@ -200,10 +226,22 @@ static SQLRETURN extract_sql_error_field_w( EHEAD *head,
         {
             SQLRETURN ret;
 
-            if ( head -> handle_type != SQL_HANDLE_STMT )
+            if ( rec_number > 0 )
             {
                 return SQL_ERROR;
             }
+            else if ( head -> handle_type != SQL_HANDLE_STMT )
+            {
+                if ( diag_info_ptr )
+                {
+                    wcscpy( diag_info_ptr, L"" );
+                }
+                if ( string_length_ptr )
+                {
+                    *string_length_ptr = 0;
+                }
+                return SQL_SUCCESS;
+            }
             else if ( head -> header_set )
             {
                 if ( SQL_SUCCEEDED( head -> diag_dynamic_function_ret ) && diag_info_ptr )
@@ -211,7 +249,7 @@ static SQLRETURN extract_sql_error_field_w( EHEAD *head,
                     wide_strncpy( diag_info_ptr, head -> diag_dynamic_function, buffer_length );
                     if ( string_length_ptr )
                     {
-                        *string_length_ptr = wide_strlen( head -> diag_dynamic_function );
+                        *string_length_ptr = wide_strlen( head -> diag_dynamic_function ) * sizeof( SQLWCHAR );
                     }
                 }
                 return head -> diag_dynamic_function_ret;
@@ -230,58 +268,6 @@ static SQLRETURN extract_sql_error_field_w( EHEAD *head,
 
                 return ret;
             }
-            else if ( !__get_connection( head ) -> unicode_driver &&
-                CHECK_SQLGETDIAGFIELD( __get_connection( head )))
-            {
-                ret = SQLGETDIAGFIELDW( __get_connection( head ),
-                        SQL_HANDLE_STMT,
-                        __get_driver_handle( head ),
-                        0,
-                        diag_identifier,
-                        diag_info_ptr,
-                        buffer_length,
-                        string_length_ptr );
-
-                return ret;
-            }
-            if ( diag_info_ptr )
-            {
-                strcpy( diag_info_ptr, "" );
-            }
-        }
-        return SQL_SUCCESS;
-
-      case SQL_DIAG_DYNAMIC_FUNCTION_CODE:
-        {
-            SQLINTEGER val;
-            SQLRETURN ret;
-
-            if ( head -> handle_type != SQL_HANDLE_STMT )
-            {
-                return SQL_ERROR;
-            }
-            else if ( head -> header_set )
-            {
-                if ( SQL_SUCCEEDED( head -> diag_dynamic_function_code_ret ) && diag_info_ptr )
-                {
-                    *((SQLINTEGER*)diag_info_ptr) = head -> diag_dynamic_function_code;
-                }
-                return head -> diag_dynamic_function_code_ret;
-            }
-            else if ( __get_connection( head ) -> unicode_driver &&
-                CHECK_SQLGETDIAGFIELDW( __get_connection( head )))
-            {
-                ret = SQLGETDIAGFIELDW( __get_connection( head ),
-                        SQL_HANDLE_STMT,
-                        __get_driver_handle( head ),
-                        0,
-                        diag_identifier,
-                        diag_info_ptr,
-                        buffer_length,
-                        string_length_ptr );
-
-                return ret;
-            }
             else if ( !__get_connection( head ) -> unicode_driver &&
                 CHECK_SQLGETDIAGFIELD( __get_connection( head )))
             {
@@ -303,7 +289,11 @@ static SQLRETURN extract_sql_error_field_w( EHEAD *head,
 
                 if ( SQL_SUCCEEDED( ret ) && as1 && diag_info_ptr )
                 {
-                    ansi_to_unicode_copy( diag_info_ptr, (char*) as1, SQL_NTS, __get_connection( head ));
+                    ansi_to_unicode_copy( diag_info_ptr, (char*) as1, SQL_NTS, __get_connection( head ), NULL);
+                }
+                if ( SQL_SUCCEEDED( ret ) && string_length_ptr )
+                {
+                    *string_length_ptr *= sizeof( SQLWCHAR );
                 }
 
                 if ( as1 )
@@ -313,6 +303,63 @@ static SQLRETURN extract_sql_error_field_w( EHEAD *head,
 
                 return ret;
             }
+            if ( diag_info_ptr )
+            {
+                strcpy( diag_info_ptr, "" );
+            }
+        }
+        return SQL_SUCCESS;
+
+      case SQL_DIAG_DYNAMIC_FUNCTION_CODE:
+        {
+            SQLINTEGER val;
+            SQLRETURN ret;
+
+            if ( rec_number > 0  )
+            {
+                return SQL_ERROR;
+            }
+            else if ( head -> handle_type != SQL_HANDLE_STMT )
+            {
+                *((SQLINTEGER*)diag_info_ptr) = 0;
+                return SQL_SUCCESS;
+            }
+            else if ( head -> header_set )
+            {
+                if ( SQL_SUCCEEDED( head -> diag_dynamic_function_code_ret ) && diag_info_ptr )
+                {
+                    *((SQLINTEGER*)diag_info_ptr) = head -> diag_dynamic_function_code;
+                }
+                return head -> diag_dynamic_function_code_ret;
+            }
+            else if ( __get_connection( head ) -> unicode_driver &&
+                CHECK_SQLGETDIAGFIELDW( __get_connection( head )))
+            {
+                ret = SQLGETDIAGFIELDW( __get_connection( head ),
+                        SQL_HANDLE_STMT,
+                        __get_driver_handle( head ),
+                        0,
+                        diag_identifier,
+                        diag_info_ptr,
+                        buffer_length,
+                        string_length_ptr );
+
+                return ret;
+            }
+            else if ( !__get_connection( head ) -> unicode_driver &&
+                CHECK_SQLGETDIAGFIELD( __get_connection( head )))
+            {
+                ret = SQLGETDIAGFIELD( __get_connection( head ),
+                        SQL_HANDLE_STMT,
+                        __get_driver_handle( head ),
+                        0,
+                        diag_identifier,
+                        diag_info_ptr,
+                        buffer_length,
+                        string_length_ptr );
+
+                return ret;
+            }
             else
             {
                 val = SQL_DIAG_UNKNOWN_STATEMENT;
@@ -329,6 +376,10 @@ static SQLRETURN extract_sql_error_field_w( EHEAD *head,
         {
             SQLINTEGER val;
             
+            if ( rec_number > 0 )
+            {
+                return SQL_ERROR;
+            }
             val = head -> sql_diag_head.internal_count + 
                 head -> sql_diag_head.error_count;
 
@@ -354,7 +405,9 @@ static SQLRETURN extract_sql_error_field_w( EHEAD *head,
      * else check the records
      */
 
-    if ( rec_number < 1 )
+    if ( rec_number < 1 ||
+        ( diag_identifier == SQL_DIAG_COLUMN_NUMBER ||
+          diag_identifier == SQL_DIAG_ROW_NUMBER ) && head -> handle_type != SQL_HANDLE_STMT )
     {
         return SQL_ERROR;
     }
@@ -413,7 +466,7 @@ static SQLRETURN extract_sql_error_field_w( EHEAD *head,
             SQLRETURN ret;
             SQLCHAR *as1 = NULL;
 
-            if ( diag_identifier == SQL_DIAG_SQLSTATE && diag_info_ptr && buffer_length > 0 )
+            if ( is_char_diag( diag_identifier ) && diag_info_ptr && buffer_length > 0 )
             {
                 as1 = malloc( buffer_length + 1 );
             }
@@ -436,7 +489,7 @@ static SQLRETURN extract_sql_error_field_w( EHEAD *head,
                 if ( diag_info_ptr && as1 )
                 {
                     __map_error_state( (char*) as1, __get_version( head ));
-                    ansi_to_unicode_copy( diag_info_ptr, (char*) as1, SQL_NTS, __get_connection( head ));
+                    ansi_to_unicode_copy( diag_info_ptr, (char*) as1, SQL_NTS, __get_connection( head ), NULL );
                 }
             }
 
@@ -481,7 +534,7 @@ static SQLRETURN extract_sql_error_field_w( EHEAD *head,
                 wide_strncpy( diag_info_ptr, ptr -> diag_class_origin, buffer_length );
                 if ( string_length_ptr )
                 {
-                    *string_length_ptr = wide_strlen( ptr -> diag_class_origin );
+                    *string_length_ptr = wide_strlen( ptr -> diag_class_origin ) * sizeof( SQLWCHAR );
                 }
                 return ptr -> diag_class_origin_ret;
             }
@@ -509,7 +562,7 @@ static SQLRETURN extract_sql_error_field_w( EHEAD *head,
                 wide_strcpy( diag_info_ptr, ptr -> diag_connection_name );
                 if ( string_length_ptr )
                 {
-                    *string_length_ptr = wide_strlen( ptr -> diag_connection_name );
+                    *string_length_ptr = wide_strlen( ptr -> diag_connection_name ) * sizeof( SQLWCHAR );
                 }
                 return ptr -> diag_connection_name_ret;
             }
@@ -542,10 +595,9 @@ static SQLRETURN extract_sql_error_field_w( EHEAD *head,
             }
             if ( string_length_ptr )
             {
-                *string_length_ptr = wide_strlen( str );
+                *string_length_ptr = wide_strlen( str ) * sizeof( SQLWCHAR );
             }
 
-            free( str );
             return ret;
         }
         break;
@@ -577,7 +629,7 @@ static SQLRETURN extract_sql_error_field_w( EHEAD *head,
                 wide_strcpy( diag_info_ptr, ptr -> diag_server_name );
                 if ( string_length_ptr )
                 {
-                    *string_length_ptr = wide_strlen( ptr -> diag_server_name );
+                    *string_length_ptr = wide_strlen( ptr -> diag_server_name ) * sizeof( SQLWCHAR );
                 }
                 return ptr -> diag_server_name_ret;
             }
@@ -617,7 +669,7 @@ static SQLRETURN extract_sql_error_field_w( EHEAD *head,
             }
             if ( string_length_ptr )
             {
-                *string_length_ptr = wide_strlen( str );
+                *string_length_ptr = wide_strlen( str ) * sizeof( SQLWCHAR );
             }
             return ret;
         }
@@ -630,7 +682,7 @@ static SQLRETURN extract_sql_error_field_w( EHEAD *head,
                 wide_strcpy( diag_info_ptr, ptr -> diag_subclass_origin );
                 if ( string_length_ptr )
                 {
-                    *string_length_ptr = wide_strlen( ptr -> diag_subclass_origin );
+                    *string_length_ptr = wide_strlen( ptr -> diag_subclass_origin ) * sizeof( SQLWCHAR );
                 }
                 return ptr -> diag_subclass_origin_ret;
             }
