@@ -4,7 +4,7 @@
  * (pharvey@codebydesign.com).
  *
  * Modified and extended by Nick Gorham
- * (nick@easysoft.com).
+ * (nick@lurcher.org).
  *
  * Any bugs or problems should be considered the fault of Nick and not
  * Peter.
@@ -112,7 +112,7 @@ static SQLRETURN extract_sql_error_rec_w( EHEAD *head,
     {
         SQLWCHAR *tmp;
 
-        tmp = ansi_to_unicode_alloc((SQLCHAR*) "00000", SQL_NTS, __get_connection( head ));
+        tmp = ansi_to_unicode_alloc((SQLCHAR*) "00000", SQL_NTS, __get_connection( head ), NULL );
         wide_strcpy( sqlstate, tmp );
         free( tmp );
     }
@@ -174,12 +174,12 @@ static SQLRETURN extract_sql_error_rec_w( EHEAD *head,
          */
 
         if ( SQL_SUCCEEDED( ret ) && sqlstate )
-            __map_error_state((char*) sqlstate, __get_version( head ));
+            __map_error_state_w(sqlstate, __get_version( head ));
 
         return ret;
     }
-    else if ( rec_number <= head -> sql_diag_head.internal_count + 
-            head -> sql_diag_head.error_count )
+    else if ( !__is_env( head ) && __get_connection( head ) -> state != STATE_C2
+        && head->sql_diag_head.error_count )
     {
         ERROR *ptr;
         rec_number -= head -> sql_diag_head.internal_count;
@@ -243,7 +243,7 @@ static SQLRETURN extract_sql_error_rec_w( EHEAD *head,
                 {
                     if ( as1 )
                     {
-                        ansi_to_unicode_copy( sqlstate,(char*) as1, SQL_NTS, __get_connection( head ));
+                        ansi_to_unicode_copy( sqlstate,(char*) as1, SQL_NTS, __get_connection( head ), NULL );
                         __map_error_state_w( sqlstate, __get_version( head ));
                     }
                 }
@@ -251,7 +251,7 @@ static SQLRETURN extract_sql_error_rec_w( EHEAD *head,
                 {
                     if ( as2 )
                     {
-                        ansi_to_unicode_copy( message_text,(char*) as2, SQL_NTS, __get_connection( head ));
+                        ansi_to_unicode_copy( message_text,(char*) as2, SQL_NTS, __get_connection( head ), NULL );
                     }
                 }
             }
@@ -339,6 +339,7 @@ SQLRETURN SQLGetDiagRecW( SQLSMALLINT handle_type,
     SQLRETURN ret;
     SQLCHAR s0[ 32 ], s1[ 100 + LOG_MESSAGE_LEN ];
     SQLCHAR s2[ 100 + LOG_MESSAGE_LEN ];
+    SQLCHAR s3[ 100 + LOG_MESSAGE_LEN ];
 
     if ( rec_number < 1 )
     {
@@ -408,13 +409,18 @@ SQLRETURN SQLGetDiagRecW( SQLSMALLINT handle_type,
 \n\t\t\tNative = %s\
 \n\t\t\tMessage Text = %s",
                         __get_return_status( ret, s2 ),
-                        ts1 = unicode_to_ansi_alloc( sqlstate, SQL_NTS, NULL ),
+                        __sdata_as_string( s3, SQL_CHAR,
+				NULL, ts1 = unicode_to_ansi_alloc( sqlstate, SQL_NTS, NULL, NULL  )),
                         __iptr_as_string( s0, native ),
                         __sdata_as_string( s1, SQL_CHAR, 
-                            text_length_ptr, ts2 = unicode_to_ansi_alloc( message_text, SQL_NTS, NULL )));
+                            text_length_ptr, ts2 = unicode_to_ansi_alloc( message_text, SQL_NTS, NULL, NULL  )));
 
-                free( ts1 );
-                free( ts2 );
+		if ( ts1 ) {
+                	free( ts1 );
+		}
+		if ( ts2 ) {
+                	free( ts2 );
+		}
             }
             else
             {
@@ -531,13 +537,18 @@ SQLRETURN SQLGetDiagRecW( SQLSMALLINT handle_type,
 \n\t\t\tNative = %s\
 \n\t\t\tMessage Text = %s",
                         __get_return_status( ret, s2 ),
-                        ts1 = unicode_to_ansi_alloc( sqlstate, SQL_NTS, connection ),
+                        __sdata_as_string( s3, SQL_CHAR,
+				NULL, ts1 = unicode_to_ansi_alloc( sqlstate, SQL_NTS, connection, NULL  )),
                         __iptr_as_string( s0, native ),
                         __sdata_as_string( s1, SQL_CHAR, 
-                            text_length_ptr, ts2 = unicode_to_ansi_alloc( message_text, SQL_NTS, connection )));
+                            	text_length_ptr, ts2 = unicode_to_ansi_alloc( message_text, SQL_NTS, connection, NULL  )));
 
-                free( ts1 );
-                free( ts2 );
+		if ( ts1 ) {
+                	free( ts1 );
+		}
+		if ( ts2 ) {
+                	free( ts2 );
+		}
             }
             else
             {
@@ -654,13 +665,18 @@ SQLRETURN SQLGetDiagRecW( SQLSMALLINT handle_type,
 \n\t\t\tNative = %s\
 \n\t\t\tMessage Text = %s",
                         __get_return_status( ret, s2 ),
-                        ts1 = unicode_to_ansi_alloc( sqlstate, SQL_NTS, statement -> connection ),
+                        __sdata_as_string( s3, SQL_CHAR,
+				NULL, ts1 = unicode_to_ansi_alloc( sqlstate, SQL_NTS, statement -> connection, NULL  )),
                         __iptr_as_string( s0, native ),
                         __sdata_as_string( s1, SQL_CHAR, 
-                            text_length_ptr, ts2 = unicode_to_ansi_alloc( message_text, SQL_NTS, statement -> connection )));
+                            text_length_ptr, ts2 = unicode_to_ansi_alloc( message_text, SQL_NTS, statement -> connection, NULL  )));
 
-                free( ts1 );
-                free( ts2 );
+		if ( ts1 ) {
+                	free( ts1 );
+		}
+		if ( ts2 ) {
+                	free( ts2 );
+		}
             }
             else
             {
@@ -777,13 +793,18 @@ SQLRETURN SQLGetDiagRecW( SQLSMALLINT handle_type,
 \n\t\t\tNative = %s\
 \n\t\t\tMessage Text = %s",
                         __get_return_status( ret, s2 ),
-                        ts1 = unicode_to_ansi_alloc( sqlstate, SQL_NTS, descriptor -> connection ),
+                        __sdata_as_string( s3, SQL_CHAR,
+				NULL, ts1 = unicode_to_ansi_alloc( sqlstate, SQL_NTS, descriptor -> connection, NULL  )),
                         __iptr_as_string( s0, native ),
                         __sdata_as_string( s1, SQL_CHAR, 
-                            text_length_ptr, ts2 = unicode_to_ansi_alloc( message_text, SQL_NTS, descriptor -> connection )));
+                            text_length_ptr, ts2 = unicode_to_ansi_alloc( message_text, SQL_NTS, descriptor -> connection, NULL  )));
 
-                free( ts1 );
-                free( ts2 );
+		if ( ts1 ) {
+                	free( ts1 );
+		}
+		if ( ts2 ) {
+                	free( ts2 );
+		}
             }
             else
             {
