@@ -4,7 +4,7 @@
  * (pharvey@codebydesign.com).
  *
  * Modified and extended by Nick Gorham
- * (nick@easysoft.com).
+ * (nick@lurcher.org).
  *
  * Any bugs or problems should be considered the fault of Nick and not
  * Peter.
@@ -101,6 +101,7 @@ SQLRETURN SQLSetDescFieldW( SQLHDESC descriptor_handle,
     DMHDESC descriptor = (DMHDESC) descriptor_handle;
     SQLRETURN ret;
     SQLCHAR s1[ 100 + LOG_MESSAGE_LEN ];
+    int isStrField = 0;
 
     /*
      * check descriptor
@@ -185,7 +186,7 @@ SQLRETURN SQLSetDescFieldW( SQLHDESC descriptor_handle,
                 ERROR_HY010, NULL,
                 descriptor -> connection -> environment -> requested_version );
 
-        return function_return( SQL_HANDLE_DESC, descriptor, SQL_ERROR );
+        return function_return_nodrv( SQL_HANDLE_DESC, descriptor, SQL_ERROR );
     }
 
     /*
@@ -196,7 +197,10 @@ SQLRETURN SQLSetDescFieldW( SQLHDESC descriptor_handle,
         __check_stmt_from_desc( descriptor, STATE_S9 ) ||
         __check_stmt_from_desc( descriptor, STATE_S10 ) ||
         __check_stmt_from_desc( descriptor, STATE_S11 ) ||
-        __check_stmt_from_desc( descriptor, STATE_S12 )) {
+        __check_stmt_from_desc( descriptor, STATE_S12 ) ||
+        __check_stmt_from_desc( descriptor, STATE_S13 ) ||
+        __check_stmt_from_desc( descriptor, STATE_S14 ) ||
+        __check_stmt_from_desc( descriptor, STATE_S15 )) {
 
         dm_log_write( __FILE__, 
                 __LINE__, 
@@ -208,7 +212,100 @@ SQLRETURN SQLSetDescFieldW( SQLHDESC descriptor_handle,
                 ERROR_HY010, NULL,
                 descriptor -> connection -> environment -> requested_version );
 
-        return function_return( SQL_HANDLE_DESC, descriptor, SQL_ERROR );
+        return function_return_nodrv( SQL_HANDLE_DESC, descriptor, SQL_ERROR );
+    }
+
+    if ( rec_number < 0 )
+    {
+        __post_internal_error( &descriptor -> error,
+                ERROR_07009, NULL,
+                descriptor -> connection -> environment -> requested_version );
+
+        return function_return_nodrv( SQL_HANDLE_DESC, descriptor, SQL_ERROR );
+    }
+
+    switch ( field_identifier )
+    {
+    /* Fixed-length fields: buffer_length is ignored */
+    case SQL_DESC_ALLOC_TYPE:
+    case SQL_DESC_ARRAY_SIZE:
+    case SQL_DESC_ARRAY_STATUS_PTR:
+    case SQL_DESC_BIND_OFFSET_PTR:
+    case SQL_DESC_BIND_TYPE:
+    case SQL_DESC_COUNT:
+    case SQL_DESC_ROWS_PROCESSED_PTR:
+    case SQL_DESC_AUTO_UNIQUE_VALUE:
+    case SQL_DESC_CASE_SENSITIVE:
+    case SQL_DESC_CONCISE_TYPE:
+    case SQL_DESC_DATA_PTR:
+    case SQL_DESC_DATETIME_INTERVAL_CODE:
+    case SQL_DESC_DATETIME_INTERVAL_PRECISION:
+    case SQL_DESC_DISPLAY_SIZE:
+    case SQL_DESC_FIXED_PREC_SCALE:
+    case SQL_DESC_INDICATOR_PTR:
+    case SQL_DESC_LENGTH:
+    case SQL_DESC_NULLABLE:
+    case SQL_DESC_NUM_PREC_RADIX:
+    case SQL_DESC_OCTET_LENGTH:
+    case SQL_DESC_OCTET_LENGTH_PTR:
+    case SQL_DESC_PARAMETER_TYPE:
+    case SQL_DESC_PRECISION:
+    case SQL_DESC_ROWVER:
+    case SQL_DESC_SCALE:
+    case SQL_DESC_SEARCHABLE:
+    case SQL_DESC_TYPE:
+    case SQL_DESC_UNNAMED:
+    case SQL_DESC_UNSIGNED:
+    case SQL_DESC_UPDATABLE:
+        isStrField = 0;
+        break;
+    /* Pointer to data: buffer_length must be valid */
+    case SQL_DESC_BASE_COLUMN_NAME:
+    case SQL_DESC_BASE_TABLE_NAME:
+    case SQL_DESC_CATALOG_NAME:
+    case SQL_DESC_LABEL:
+    case SQL_DESC_LITERAL_PREFIX:
+    case SQL_DESC_LITERAL_SUFFIX:
+    case SQL_DESC_LOCAL_TYPE_NAME:
+    case SQL_DESC_NAME:
+    case SQL_DESC_SCHEMA_NAME:
+    case SQL_DESC_TABLE_NAME:
+    case SQL_DESC_TYPE_NAME:
+        isStrField = 1;
+        break;
+    default:
+        isStrField = buffer_length != SQL_IS_POINTER && buffer_length != SQL_IS_INTEGER
+            && buffer_length != SQL_IS_UINTEGER && buffer_length != SQL_IS_SMALLINT &&
+            buffer_length != SQL_IS_USMALLINT;
+    }
+    
+    if ( isStrField && buffer_length < 0 && buffer_length != SQL_NTS)
+    {
+        __post_internal_error( &descriptor -> error,
+            ERROR_HY090, NULL,
+            descriptor -> connection -> environment -> requested_version );
+
+        return function_return_nodrv( SQL_HANDLE_DESC, descriptor, SQL_ERROR );
+    }
+
+    if ( field_identifier == SQL_DESC_COUNT && (SQLINTEGER)value < 0 )
+    {
+        __post_internal_error( &descriptor -> error,
+                ERROR_07009, NULL,
+                descriptor -> connection -> environment -> requested_version );
+
+        return function_return_nodrv( SQL_HANDLE_DESC, descriptor, SQL_ERROR );
+    }
+
+    if ( field_identifier == SQL_DESC_PARAMETER_TYPE && value != SQL_PARAM_INPUT
+        && value != SQL_PARAM_OUTPUT && value != SQL_PARAM_INPUT_OUTPUT &&
+        value != SQL_PARAM_INPUT_OUTPUT_STREAM && value != SQL_PARAM_OUTPUT_STREAM )
+    {
+        __post_internal_error( &descriptor -> error,
+                ERROR_HY105, NULL,
+                descriptor -> connection -> environment -> requested_version );
+
+        return function_return_nodrv( SQL_HANDLE_DESC, descriptor, SQL_ERROR );
     }
 
     if ( descriptor -> connection -> unicode_driver ||
@@ -226,7 +323,7 @@ SQLRETURN SQLSetDescFieldW( SQLHDESC descriptor_handle,
                 ERROR_IM001, NULL,
                 descriptor -> connection -> environment -> requested_version );
 
-        	return function_return( SQL_HANDLE_DESC, descriptor, SQL_ERROR );
+        	return function_return_nodrv( SQL_HANDLE_DESC, descriptor, SQL_ERROR );
 		}
 
     	ret = SQLSETDESCFIELDW( descriptor -> connection,
@@ -265,7 +362,7 @@ SQLRETURN SQLSetDescFieldW( SQLHDESC descriptor_handle,
                 ERROR_IM001, NULL,
                 descriptor -> connection -> environment -> requested_version );
 
-        	return function_return( SQL_HANDLE_DESC, descriptor, SQL_ERROR );
+        	return function_return_nodrv( SQL_HANDLE_DESC, descriptor, SQL_ERROR );
 		}
 
 		/*
@@ -275,7 +372,7 @@ SQLRETURN SQLSetDescFieldW( SQLHDESC descriptor_handle,
 		switch ( field_identifier )
 		{
 			case SQL_DESC_NAME:		/* This is the only R/W SQLCHAR* type */
-        		ascii_str = (SQLCHAR*) unicode_to_ansi_alloc( value, buffer_length, descriptor -> connection );
+        		ascii_str = (SQLCHAR*) unicode_to_ansi_alloc( value, buffer_length, descriptor -> connection, NULL );
 				value = ascii_str;
 				buffer_length = strlen((char*) ascii_str );
 				break;
