@@ -27,6 +27,7 @@
 #endif
 
 static int OpenDatabase( SQLHENV *phEnv, SQLHDBC *phDbc, char *szDSN, char *szUID, char *szPWD );
+static int CloseDatabase( SQLHENV hEnv, SQLHDBC hDbc );
 static int ExecuteSQL( SQLHDBC hDbc, char *szSQL, char cDelimiter, int bColumnNames, int bHTMLTable );
 static int ExecuteHelp( SQLHDBC hDbc, char *szSQL, char cDelimiter, int bColumnNames, int bHTMLTable );
 
@@ -48,6 +49,7 @@ static void output_help(void);
 int     bVerbose                    = 0;
 SQLHENV hEnv                        = 0;
 SQLHDBC hDbc                        = 0;
+int     buseED                      = 0;
 
 void UWriteHeaderNormal( SQLHSTMT hStmt, SQLTCHAR *szSepLine );
 void UWriteFooterNormal( SQLHSTMT hStmt, SQLTCHAR *szSepLine, SQLLEN nRows );
@@ -137,6 +139,9 @@ int main( int argc, char *argv[] )
                 case 'v':
                     bVerbose = 1;
                     break;
+                case 'e':
+                    buseED = 1;
+                    break;
                 case '-':
                     printf( "unixODBC " VERSION "\n" );
                     exit(0);
@@ -155,7 +160,7 @@ int main( int argc, char *argv[] )
                     break;
 #endif
                 default:
-                    strcpy( stderr, szSyntax );
+                    fputs( szSyntax, stderr );
                     exit( 1 );
             }
             continue;
@@ -268,7 +273,7 @@ int main( int argc, char *argv[] )
 /****************************
  * OpenDatabase - do everything we have to do to get a viable connection to szDSN
  ***************************/
-int OpenDatabase( SQLHENV *phEnv, SQLHDBC *phDbc, char *szDSN, char *szUID, char *szPWD )
+static int OpenDatabase( SQLHENV *phEnv, SQLHDBC *phDbc, char *szDSN, char *szUID, char *szPWD )
 {
     SQLCHAR dsn[ 1024 ], uid[ 1024 ], pwd[ 1024 ];
     SQLTCHAR cstr[ 1024 ];
@@ -349,6 +354,7 @@ int OpenDatabase( SQLHENV *phEnv, SQLHDBC *phDbc, char *szDSN, char *szUID, char
     {
         cstr[ i ] = zcstr[ i ];
     }
+    cstr[ i ] = 0;
 
     if ( !SQL_SUCCEEDED( SQLDriverConnect( *phDbc, NULL, cstr, SQL_NTS, NULL, 0, NULL, SQL_DRIVER_NOPROMPT  )))
     {
@@ -367,7 +373,7 @@ int OpenDatabase( SQLHENV *phEnv, SQLHDBC *phDbc, char *szDSN, char *szUID, char
  * ExecuteSQL - create a statement, execute the SQL, and get rid of the statement
  *            - show results as per request; bHTMLTable has precedence over other options
  ***************************/
-int ExecuteSQL( SQLHDBC hDbc, char *szSQL, char cDelimiter, int bColumnNames, int bHTMLTable )
+static int ExecuteSQL( SQLHDBC hDbc, char *szSQL, char cDelimiter, int bColumnNames, int bHTMLTable )
 {
     SQLHSTMT        hStmt;
     SQLTCHAR        szSepLine[32001];   
@@ -390,31 +396,54 @@ int ExecuteSQL( SQLHDBC hDbc, char *szSQL, char cDelimiter, int bColumnNames, in
         return 0;
     }
 
-    if ( SQLPrepare( hStmt, szUcSQL, SQL_NTS ) != SQL_SUCCESS )
-    {
-        if ( bVerbose ) DumpODBCLog( hEnv, hDbc, hStmt );
-        fprintf( stderr, "[ISQL]ERROR: Could not SQLPrepare\n" );
-        SQLFreeStmt( hStmt, SQL_DROP );
-        return 0;
-    }
+    if ( buseED ) {
+        ret = SQLExecDirect( hStmt, szUcSQL, SQL_NTS );
 
-    ret =  SQLExecute( hStmt );
-
-    if ( ret == SQL_NO_DATA )
-    {
-        fprintf( stderr, "[ISQL]INFO: SQLExecute returned SQL_NO_DATA\n" );
+        if ( ret == SQL_NO_DATA )
+        {
+            fprintf( stderr, "[ISQL]INFO: SQLExecDirect returned SQL_NO_DATA\n" );
+        }
+        else if ( ret == SQL_SUCCESS_WITH_INFO )
+        {
+            if ( bVerbose ) DumpODBCLog( hEnv, hDbc, hStmt );
+            fprintf( stderr, "[ISQL]INFO: SQLExecDirect returned SQL_SUCCESS_WITH_INFO\n" );
+        }
+        else if ( ret != SQL_SUCCESS )
+        {
+            if ( bVerbose ) DumpODBCLog( hEnv, hDbc, hStmt );
+            fprintf( stderr, "[ISQL]ERROR: Could not SQLExecDirect\n" );
+            SQLFreeStmt( hStmt, SQL_DROP );
+            free(szSepLine);
+            return 0;
+        }
     }
-    else if ( ret == SQL_SUCCESS_WITH_INFO )
-    {
-        if ( bVerbose ) DumpODBCLog( hEnv, hDbc, hStmt );
-        fprintf( stderr, "[ISQL]INFO: SQLExecute returned SQL_SUCCESS_WITH_INFO\n" );
-    }
-    else if ( ret != SQL_SUCCESS )
-    {
-        if ( bVerbose ) DumpODBCLog( hEnv, hDbc, hStmt );
-        fprintf( stderr, "[ISQL]ERROR: Could not SQLExecute\n" );
-        SQLFreeStmt( hStmt, SQL_DROP );
-        return 0;
+    else {
+        if ( SQLPrepare( hStmt, szUcSQL, SQL_NTS ) != SQL_SUCCESS )
+        {
+            if ( bVerbose ) DumpODBCLog( hEnv, hDbc, hStmt );
+            fprintf( stderr, "[ISQL]ERROR: Could not SQLPrepare\n" );
+            SQLFreeStmt( hStmt, SQL_DROP );
+            return 0;
+        }
+    
+        ret =  SQLExecute( hStmt );
+    
+        if ( ret == SQL_NO_DATA )
+        {
+            fprintf( stderr, "[ISQL]INFO: SQLExecute returned SQL_NO_DATA\n" );
+        }
+        else if ( ret == SQL_SUCCESS_WITH_INFO )
+        {
+            if ( bVerbose ) DumpODBCLog( hEnv, hDbc, hStmt );
+            fprintf( stderr, "[ISQL]INFO: SQLExecute returned SQL_SUCCESS_WITH_INFO\n" );
+        }
+        else if ( ret != SQL_SUCCESS )
+        {
+            if ( bVerbose ) DumpODBCLog( hEnv, hDbc, hStmt );
+            fprintf( stderr, "[ISQL]ERROR: Could not SQLExecute\n" );
+            SQLFreeStmt( hStmt, SQL_DROP );
+            return 0;
+        }
     }
 
     do 
@@ -476,7 +505,7 @@ int ExecuteSQL( SQLHDBC hDbc, char *szSQL, char cDelimiter, int bColumnNames, in
  * ExecuteHelp - create a statement, execute the SQL, and get rid of the statement
  *             - show results as per request; bHTMLTable has precedence over other options
  ***************************/
-int ExecuteHelp( SQLHDBC hDbc, char *szSQL, char cDelimiter, int bColumnNames, int bHTMLTable )
+static int ExecuteHelp( SQLHDBC hDbc, char *szSQL, char cDelimiter, int bColumnNames, int bHTMLTable )
 {
     char            szTable[250]                        = "";
     SQLHSTMT        hStmt;
@@ -574,7 +603,7 @@ int CloseDatabase( SQLHENV hEnv, SQLHDBC hDbc )
 /****************************
  * WRITE HTML
  ***************************/
-void WriteHeaderHTMLTable( SQLHSTMT hStmt )
+static void WriteHeaderHTMLTable( SQLHSTMT hStmt )
 {
     SQLINTEGER      nCol                            = 0;
     SQLSMALLINT     nColumns                        = 0;
@@ -601,7 +630,7 @@ void WriteHeaderHTMLTable( SQLHSTMT hStmt )
     printf( "</tr>\n" );
 }
 
-void WriteBodyHTMLTable( SQLHSTMT hStmt )
+static void WriteBodyHTMLTable( SQLHSTMT hStmt )
 {
     SQLINTEGER      nCol                            = 0;
     SQLSMALLINT     nColumns                        = 0;
@@ -647,7 +676,7 @@ void WriteBodyHTMLTable( SQLHSTMT hStmt )
     }
 }
 
-void WriteFooterHTMLTable( SQLHSTMT hStmt )
+static void WriteFooterHTMLTable( SQLHSTMT hStmt )
 {
     printf( "</table>\n" );
 }
@@ -658,7 +687,7 @@ void WriteFooterHTMLTable( SQLHSTMT hStmt )
  * - last column no longer has a delimit char (it is implicit)...
  *   this is consistent with odbctxt
  ***************************/
-void WriteHeaderDelimited( SQLHSTMT hStmt, char cDelimiter )
+static void WriteHeaderDelimited( SQLHSTMT hStmt, char cDelimiter )
 {
     SQLINTEGER      nCol                            = 0;
     SQLSMALLINT     nColumns                        = 0;
@@ -679,7 +708,7 @@ void WriteHeaderDelimited( SQLHSTMT hStmt, char cDelimiter )
     putchar( '\n' );
 }
 
-void WriteBodyDelimited( SQLHSTMT hStmt, char cDelimiter )
+static void WriteBodyDelimited( SQLHSTMT hStmt, char cDelimiter )
 {
     SQLINTEGER      nCol                            = 0;
     SQLSMALLINT     nColumns                        = 0;
@@ -773,7 +802,7 @@ void UWriteHeaderNormal( SQLHSTMT hStmt, SQLTCHAR *szSepLine )
     puts((char*) szSepLine );
 }
 
-SQLLEN WriteBodyNormal( SQLHSTMT hStmt )
+static SQLLEN WriteBodyNormal( SQLHSTMT hStmt )
 {
     SQLINTEGER      nCol                            = 0;
     SQLSMALLINT     nColumns                        = 0;
@@ -870,7 +899,7 @@ void UWriteFooterNormal( SQLHSTMT hStmt, SQLTCHAR   *szSepLine, SQLLEN nRows )
 
 
 
-int DumpODBCLog( SQLHENV hEnv, SQLHDBC hDbc, SQLHSTMT hStmt )
+static int DumpODBCLog( SQLHENV hEnv, SQLHDBC hDbc, SQLHSTMT hStmt )
 {
     SQLTCHAR        szError[501];
     SQLTCHAR        szSqlState[10];
