@@ -8,6 +8,7 @@
  * -----------------------------------------------
  * Peter Harvey		- pharvey@codebydesign.com
  **************************************************/
+#include <config.h>
 #include <odbcinstext.h>
 
 BOOL SQLWriteFileDSN(			LPCSTR	pszFileName,
@@ -17,17 +18,17 @@ BOOL SQLWriteFileDSN(			LPCSTR	pszFileName,
 {
 	HINI	hIni;
 	char	szFileName[ODBC_FILENAME_MAX+1];
-	char	szFilePath[ODBC_FILENAME_MAX+1];
 
 	if ( pszFileName[0] == '/' )
 	{
-		strcpy( szFileName, pszFileName );
+		strncpy( szFileName, sizeof(szFileName) - 5, pszFileName );
 	}
 	else
 	{	
-        sprintf( szFileName, "%s/ODBCDataSources", odbcinst_system_file_path());
-        SQLGetPrivateProfileString( "ODBC", "FILEDSNPATH", szFileName, szFilePath, sizeof( szFilePath ), "odbcinst.ini" );
-        sprintf( szFileName, "%s/%s", szFilePath, pszFileName );
+		char szPath[ODBC_FILENAME_MAX+1];
+		*szPath = '\0';
+		_odbcinst_FileINI( szPath );
+		snprintf( szFileName, sizeof(szFileName) - 5, "%s/%s", szPath, pszFileName );
 	}
 
     if ( strlen( szFileName ) < 4 || strcmp( szFileName + strlen( szFileName ) - 4, ".dsn" ))
